// bcc32 -q -O1 -v- -c ccolor.cpp
// ilink32 -v- -Tpe -aa -c c0w32.obj ccolor.obj,ccolor.exe,,import32.lib comdlg32.lib shlwapi.lib cw32.lib,,ccolor.res
/******************************************************************************
 *    ccolor.cpp
 *
 *    This file is part of CColor Converter
 *    Copyright (C) 2002 Tom N Harris
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/
#define _WINVER 	0x0400
#define _WIN32_IE	0x0400
#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <commdlg.h>
#include <shlwapi.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <stdio.h>

#include "ccolor.rh"
#include "ccolor.h"


LRESULT CALLBACK WindowFunc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK DialogFunc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

void drawColor(HWND hwnd, LPDRAWITEMSTRUCT item);
void drawColorTable(HWND hwnd, LPDRAWITEMSTRUCT item);
HPALETTE LoadColorTable(HWND hwnd, HPALETTE ctable);
COLORREF pickColor(HWND hwnd);
BOOL doCTableClick(HWND hwnd, HWND item, DWORD msg);

const char szWinName[] = "Color Converter";
char gValueStr[4][32];
char gHSVorHSLStr[2][32];
char gHueStr[2][32];
HWND gMain;
static CColor *gColor;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
#pragma argsused(hInstance,nShowCmd)
	HWND hwnd, hdlg;
	MSG msg;
	WNDCLASS wcl;
	int i;

	for (i=0;i<2;i++)
		LoadString(hInstance, 1025+i, gValueStr[i], 32);
	for (i=0;i<2;i++)
		LoadString(hInstance, 1029+i, gHueStr[i], 32);
	for (i=0;i<2;i++)
		LoadString(hInstance, 1031+i, gHSVorHSLStr[i], 32);

	//define a window class
	wcl.hInstance = hInstance; //hand to this instance
	wcl.lpszClassName = szWinName; // window class name
	wcl.lpfnWndProc = WindowFunc; //window function
	wcl.style = 0;  // default style

	wcl.hIcon = LoadIcon(NULL, IDI_APPLICATION);

	wcl.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcl.lpszMenuName = NULL;  // no menu

	wcl.cbClsExtra = 0;  // no extra
	wcl.cbWndExtra = 0;
	//background
	wcl.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);

	// register the window class
	if(!RegisterClass(&wcl)) return 0;

	// lets create a window
	gMain = CreateWindow(
		szWinName,		//name of window class
		"Color Converter", 		//title
		WS_POPUP, 		//window style -normal
		CW_USEDEFAULT, 		// X Coordiate, default
		CW_USEDEFAULT, 		// Y Coordiate, default
		CW_USEDEFAULT, 		//width, default
		CW_USEDEFAULT, 		//height, default
		NULL, 		//no parent window
		NULL, 		//use menu registered with this class
		hInstance, 		//handle of this instance of the program
		NULL 		//no additional arguments
		);

	hdlg = CreateDialog(hInstance, MAKEINTRESOURCE(DIALOG_MAIN), NULL, DialogFunc);

	//Display window
	ShowWindow(hdlg, nShowCmd);
	UpdateWindow(hdlg);

	//create the message loop
	while(GetMessage(&msg, NULL, 0, 0))
	{
		if (!IsDialogMessage(hdlg, &msg))
		{
			TranslateMessage(&msg); //allow use of keyboard
			DispatchMessage(&msg);	//return control to windows
		}
	}
	return msg.wParam;
}

LRESULT CALLBACK WindowFunc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message){
	case WM_DESTROY: //terminate the program
		PostQuitMessage(0);
		break;
	default:
	//let windows do the job! you paid too much for it!!!
	return DefWindowProc(hwnd, message, wParam, lParam);
	}
	return 0;
}

BOOL CALLBACK DialogFunc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	char buf[16];
	int i;

	switch (message)
	{
	case WM_INITDIALOG:
		if ((gColor = new CColor(hwnd)) == NULL)
		{
			DestroyWindow(gMain);
			return FALSE;
		}
		for (i=0;i<2;i++)
		{
			SendDlgItemMessage(hwnd, COMBO_RGB, CB_INSERTSTRING, i, (LPARAM)gValueStr[i]);
			SendDlgItemMessage(hwnd, COMBO_HUE, CB_INSERTSTRING, i, (LPARAM)gValueStr[i]);
		}
		SendDlgItemMessage(hwnd, COMBO_RGB, CB_SETCURSEL, 0, 0);
		SendDlgItemMessage(hwnd, COMBO_HUE, CB_SETCURSEL, 0, 0);
		for (i=0;i<2;i++)
		{
			SendDlgItemMessage(hwnd, COMBO_PERCENT_HUE, CB_INSERTSTRING, i, (LPARAM)gHueStr[i]);
		}
		SendDlgItemMessage(hwnd, COMBO_PERCENT_HUE, CB_SETCURSEL, 0, 0);
		for (i=0;i<2;i++)
		{
			SendDlgItemMessage(hwnd, COMBO_HSV_OR_HSL, CB_INSERTSTRING, i, (LPARAM)gHSVorHSLStr[i]);
		}
		SendDlgItemMessage(hwnd, COMBO_HSV_OR_HSL, CB_SETCURSEL, 0, 0);
		EnableWindow(GetDlgItem(hwnd, EDIT_INDEX), FALSE);
		gColor->setColorRGB(RGB(0,0,0), 0);
		return TRUE;
break;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case STATIC_COLOR:
			//if (HIWORD(wParam) == STN_CLICKED)
			{
				return gColor->setColorRGB(pickColor(hwnd), 0);
			}
			break;
		case STATIC_CTABLE:
			if (HIWORD(wParam) == STN_CLICKED)
			{
				//UpdateDialog(hwnd, 0, gColor);
				//return TRUE;
			}
			break;
		case BUTTON_CTABLE:
			if (gColor->setColorTable(LoadColorTable(hwnd, gColor->getColorTable())))
			{
				EnableWindow(GetDlgItem(hwnd, EDIT_INDEX), TRUE);
				return TRUE;
			}
			break;
		case COMBO_RGB:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				gColor->setRGBMode(SendDlgItemMessage(hwnd, COMBO_RGB, CB_GETCURSEL, 0,0));
			}
			break;
		case COMBO_HUE:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				gColor->setHueMode(SendDlgItemMessage(hwnd, COMBO_HUE, CB_GETCURSEL, 0,0));
			}
			break;
		case COMBO_PERCENT_HUE:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				gColor->setHuePercent(SendDlgItemMessage(hwnd, COMBO_PERCENT_HUE, CB_GETCURSEL, 0,0));
			}
			break;
		case COMBO_HSV_OR_HSL:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				gColor->setHSVorHSLMode(SendDlgItemMessage(hwnd, COMBO_HSV_OR_HSL, CB_GETCURSEL, 0,0));
			}
			break;
		case EDIT_RED:
		case EDIT_GREEN:
		case EDIT_BLUE:
		case EDIT_PERCENT_RED:
		case EDIT_PERCENT_GREEN:
		case EDIT_PERCENT_BLUE:
		case EDIT_HUE:
		case EDIT_SATURATION:
		case EDIT_VALUE:
		case EDIT_PERCENT_HUE:
		case EDIT_PERCENT_SATURATION:
		case EDIT_PERCENT_VALUE:
			if (HIWORD(wParam) == EN_CHANGE)
			{
				return gColor->update(LOWORD(wParam));
			}
			break;
		case EDIT_INDEX:
			if (HIWORD(wParam) == EN_CHANGE)
			{
				GetDlgItemText(hwnd, EDIT_INDEX, buf, 16);
				gColor->setColorTableIndex(strtoul(buf, NULL, 10));
			}
			break;
		}
		break;
	case WM_DRAWITEM:
		if (((LPDRAWITEMSTRUCT)lParam)->CtlID == STATIC_COLOR)
		{
			if (((LPDRAWITEMSTRUCT)lParam)->itemAction == ODA_FOCUS)
				DrawFocusRect(((LPDRAWITEMSTRUCT)lParam)->hDC, &((LPDRAWITEMSTRUCT)lParam)->rcItem);
			else
				drawColor(hwnd, (LPDRAWITEMSTRUCT)lParam);
			return TRUE;
		}
		else if (((LPDRAWITEMSTRUCT)lParam)->CtlID == STATIC_CTABLE)
		{
			drawColorTable(hwnd, (LPDRAWITEMSTRUCT)lParam);
			return TRUE;
		}
		break;
	case WM_CLOSE:
		DestroyWindow(gMain);
		return TRUE;
	case 32:
		if ((HWND)wParam == GetDlgItem(hwnd, STATIC_CTABLE) && HIWORD(lParam) == WM_LBUTTONDOWN)
		{
			if (doCTableClick(hwnd, (HWND)wParam, HIWORD(lParam)))
			{
				return TRUE;
			}
		}
		break;
	}
	return FALSE;
}

COLORREF pickColor(HWND hwnd)
{
	static COLORREF cust[16];
	CHOOSECOLOR cc;

	memset(&cc, 0, sizeof(CHOOSECOLOR));
	cc.lStructSize = sizeof(CHOOSECOLOR);
	cc.hwndOwner = hwnd;
	cc.lpCustColors = (LPDWORD)cust;
	cc.rgbResult = gColor->getColor();
	cc.Flags = CC_ANYCOLOR | CC_FULLOPEN | CC_RGBINIT;
	if (ChooseColor(&cc))
		return cc.rgbResult;
	return gColor->getColor();
}

void drawColor(HWND hwnd, LPDRAWITEMSTRUCT item)
{
#pragma argsused(item)
	RECT	r;
	HBRUSH	b, ob;

	CopyRect(&r, &item->rcItem);
	if (item->itemState & ODS_FOCUS)
		DrawFocusRect(item->hDC, &r);
	SelectObject(item->hDC, GetStockObject(BLACK_PEN));
	b = CreateSolidBrush(gColor->getColor());
	SelectObject(item->hDC, b);
	InflateRect(&r, -2, -2);
	Rectangle(item->hDC, r.top, r.left, r.right, r.bottom);
	SelectObject(item->hDC, ob);
	DeleteObject(b);
}

void drawColorTable(HWND hwnd, LPDRAWITEMSTRUCT item)
{
#pragma argsused(item)
	HPALETTE t;
	HBRUSH b;
	PALETTEENTRY p;
	//COLORREF c;
	RECT r;
	int z, i;

	if ((t = gColor->getColorTable()) == NULL)
		return;
	i = 0;
	z = (item->rcItem.right - item->rcItem.left)/32;
	//z = 5;
	r.left = r.top = 0;
	r.right = r.bottom = z-1;
	//for (i=0;i<256;i++)
	while (i<256)
	{
		GetPaletteEntries(t, i, 1, &p);
		b = CreateSolidBrush(RGB(p.peRed,p.peGreen,p.peBlue));
		FillRect(item->hDC, &r, b);
		r.left += z;
		i++;
		//if (i%32 == 0)
		if ((r.right += z) > z*32)
		{
			r.left = 0; r.right = z-1;
			r.top += z; r.bottom += z;
		}
		DeleteObject(b);
	}
}

HPALETTE LoadColorTable(HWND hwnd, HPALETTE ctable)
{
	char szFile[260];
	OPENFILENAME ofn;
	HANDLE file;
	RIFFPal h;
	HPALETTE newtable;
	LOGPALETTE *lp;
	PALETTEENTRY *pe;
	RGBTRIPLE *p;
	DWORD count;
	char *buf, *c;
	int i;

	memset(&ofn, 0, sizeof(OPENFILENAME));
	memset(szFile, 0, sizeof(szFile));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hwnd;
	ofn.lpstrFile = szFile;
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrFilter = "All Palettes\0*.PAL;*.ACT\0MS Palette\0*.PAL\0Adobe Palette\0*.ACT\0All Files\0*.*\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
	if (!GetOpenFileName(&ofn))
	{
		return ctable;
	}

	if ((file = CreateFile(ofn.lpstrFile, GENERIC_READ, FILE_SHARE_READ, NULL, 
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE)
	{
		return ctable;
	}

	ReadFile(file, (LPVOID)&h, sizeof(h), &count, NULL);
	if (h.riffsig == 0x46464952)	// RIFF
	{
		if (h.psig1 != 0x204c4150)
		{
			CloseHandle(file);
			return ctable;
		}
		if ((lp = (LOGPALETTE*)calloc(1,h.length)) == NULL)
		{
			CloseHandle(file);
			return ctable;
		}
		ReadFile(file, lp, h.length, &count, NULL);
		CloseHandle(file);
		if (count < h.length)
		{
			free(lp);
			return ctable;
		}
	}
	else if (h.riffsig == 0x4353414a)	// JASC
	{
		SetFilePointer(file, 0, 0, FILE_BEGIN);
		if ((buf = (char*)calloc(1,3360)) == NULL)
		{
			CloseHandle(file);
			return ctable;
		}
		ReadFile(file, buf, 3352, &count, NULL);
		if (strncmp(buf, "JASC-PAL", 8))
		{
			free(buf);
			CloseHandle(file);
			return ctable;
		}
		c = strchr(buf, '\n')+1;
		c = strchr(c, '\n')+1;
		count = strtoul(c, NULL, 10);
		if ((lp = (LOGPALETTE*)calloc(1,sizeof(LOGPALETTE)+(sizeof(PALETTEENTRY)*255))) == NULL)
		{
			free(buf);
			CloseHandle(file);
			return ctable;
		}
		lp->palVersion = 0x300;
		lp->palNumEntries = count;
		pe = lp->palPalEntry;
		for (i=0;i<count;i++)
		{
			c = strchr(c, '\n')+1;
			sscanf(c, "%d %d %d", &pe[i].peRed, &pe[i].peGreen, &pe[i].peBlue);
		}
		free(buf);
		CloseHandle(file);
	}
	else {
		SetFilePointer(file, 0, 0, FILE_BEGIN);
		if ((p = (RGBTRIPLE*)calloc(256,sizeof(RGBTRIPLE))) == NULL)
		{
			CloseHandle(file);
			return ctable;
		}
		ReadFile(file, p, sizeof(RGBTRIPLE)*256, &count, NULL);
		CloseHandle(file);
		if (count < sizeof(RGBTRIPLE)*256)
		{
			free(p);
			return ctable;
		}
		if ((lp = (LOGPALETTE*)calloc(1,sizeof(LOGPALETTE)+(sizeof(PALETTEENTRY)*255))) == NULL)
		{
			free(p);
			return ctable;
		}
		lp->palVersion = 0x300;
		lp->palNumEntries = 256;
		pe = lp->palPalEntry;
		for (i=0;i<256;i++)
		{
			*((RGBTRIPLE*)(&(pe[i]))) = p[i];
		}
		free(p);
	}

	if ((newtable = CreatePalette(lp)) != NULL)
	{
		if (ctable) DeleteObject(ctable);
		ctable = newtable;
		InvalidateRect(GetDlgItem(hwnd, STATIC_CTABLE), NULL, FALSE);
	}
	free(lp);
	return ctable;
}

BOOL doCTableClick(HWND hwnd, HWND item, DWORD msg)
{
#pragma argsused(hwnd,item)
	char buf[32];
	POINT p;
	RECT r;
	int i,x,y;

	if (gColor->getColorTable() == NULL)
		return FALSE;
	GetCursorPos(&p);
	ScreenToClient(item, &p);
	GetClientRect(item, &r);
	i = (r.right - r.left)/32;
	x = p.x/i;
	y = p.y/i;
	i = y*32 + x;
	itoa(i, buf, 10);
	SetDlgItemText(hwnd, EDIT_INDEX, buf);

	return TRUE;
}

// CColor member functions

CColor::CColor()
{
	gCTable = NULL;
	gHWND = NULL;
	gColor = gRGBMode = gHueMode = gHuePercent = gHSVorHSLMode = 0;
	gUpdating = false;
}

CColor::CColor(HWND h)
{
	gCTable = NULL;
	gHWND = h;
	gColor = gRGBMode = gHueMode = gHuePercent = gHSVorHSLMode = 0;
	gUpdating = false;
}

#if 0
CColor::CColor(HWND h, COLORREF rgb)
{
	gCTable = NULL;
	gRGBMode = gHueMode = gHuePercent = gHSVorHSLMode = 0;
	gUpdating = false;
	gHWND = h;
	gColor = rgb;
	UpdateDialog(0);
}
#endif

CColor::~CColor()
{
	if (gCTable != NULL)
		DeleteObject(gCTable);
}

bool CColor::setColorTable(HPALETTE pal)
{
	if (pal != gCTable)
	{
		gCTable = pal;
		UpdateDialog();
		return true;
	}
	return false;
}

bool CColor::update(int field)
{
	char buf[16];
	//COLORREF rgb;
	RGBTRIPLE color;
	int base, i;
	double f;

	if (gUpdating)
		return false;
	
	switch (field & 0x0FF0)
	{
	case MASK_RGB:
		base = (gRGBMode == 0) ? 10 : 16;
		GetDlgItemText(gHWND, EDIT_RED, buf, 16);
		color.rgbtRed = ((i = strtoul(buf, NULL, base)) > 255) ? 255 : i;
		GetDlgItemText(gHWND, EDIT_GREEN, buf, 16);
		color.rgbtGreen = ((i = strtoul(buf, NULL, base)) > 255) ? 255 : i;
		GetDlgItemText(gHWND, EDIT_BLUE, buf, 16);
		color.rgbtBlue = ((i = strtoul(buf, NULL, base)) > 255) ? 255 : i;
		gColor = RGB(color.rgbtRed,color.rgbtGreen,color.rgbtBlue);
		break;
	case MASK_PERCENT_RGB:
		GetDlgItemText(gHWND, EDIT_PERCENT_RED, buf, 16);
		if ((f = strtod(buf, NULL)) > 100) f = 100.0;
		color.rgbtRed = static_cast<BYTE>((f*255.0)/100);
		GetDlgItemText(gHWND, EDIT_PERCENT_GREEN, buf, 16);
		if ((f = strtod(buf, NULL)) > 100) f = 100.0;
		color.rgbtGreen = static_cast<BYTE>((f*255.0)/100);
		GetDlgItemText(gHWND, EDIT_PERCENT_BLUE, buf, 16);
		if ((f = strtod(buf, NULL)) > 100) f = 100.0;
		color.rgbtBlue = static_cast<BYTE>((f*255.0)/100);
		gColor = RGB(color.rgbtRed,color.rgbtGreen,color.rgbtBlue);
		break;
	case MASK_HUE:
		base = (gHueMode == 0) ? 10 : 16;
		GetDlgItemText(gHWND, EDIT_HUE, buf, 16);
		color.rgbtRed = ((i = strtoul(buf, NULL, base)) > 240) ? 240 : i;
		GetDlgItemText(gHWND, EDIT_SATURATION, buf, 16);
		color.rgbtGreen = ((i = strtoul(buf, NULL, base)) > 240) ? 240 : i;
		GetDlgItemText(gHWND, EDIT_VALUE, buf, 16);
		color.rgbtBlue = ((i = strtoul(buf, NULL, base)) > 240) ? 240 : i;
		gColor = (gHSVorHSLMode == 0) ? hsv_to_rgb(color.rgbtRed, color.rgbtGreen, color.rgbtBlue) 
			: ColorHLSToRGB(color.rgbtRed,color.rgbtGreen,color.rgbtBlue);
		break;
	case MASK_PERCENT_HUE:
		GetDlgItemText(gHWND, EDIT_PERCENT_HUE, buf, 16);
		f = strtod(buf, NULL);
		if (gHuePercent == 0)
		{
			if (f > 360) f = 0.0;
			color.rgbtRed = static_cast<BYTE>((f*240.0)/360);
		} else {
			if (f > 100) f = 100.0;
			color.rgbtRed = static_cast<BYTE>((f*240.0)/100);
		}
		GetDlgItemText(gHWND, EDIT_PERCENT_SATURATION, buf, 16);
		f = strtod(buf, NULL);
		if (f > 100) f = 100.0;
		color.rgbtGreen = static_cast<BYTE>((f*240.0)/100);
		GetDlgItemText(gHWND, EDIT_PERCENT_VALUE, buf, 16);
		f = strtod(buf, NULL);
		if (f > 100) f = 100.0;
		color.rgbtBlue = static_cast<BYTE>((f*240.0)/100);
		gColor = ColorHLSToRGB(color.rgbtRed,color.rgbtGreen,color.rgbtBlue);
		gColor = (gHSVorHSLMode == 0) ? hsv_to_rgb(color.rgbtRed, color.rgbtGreen, color.rgbtBlue) 
			: ColorHLSToRGB(color.rgbtRed,color.rgbtGreen,color.rgbtBlue);
		break;
	default:
		return false;
	}
	UpdateDialog(field);
	return true;
}

void CColor::UpdateDialog(int excl)
{
	HWND itemhWnd;
	HDC itemDC;
	char buf[16];
	RGBTRIPLE c;
	WORD h,s,v;
	int b;

	if (!gHWND)
		return;
	
	gUpdating = true;

	itemhWnd = GetDlgItem(gHWND, STATIC_COLOR);
	itemDC = GetDC(itemhWnd);
	SetBkColor(itemDC, gColor);
	InvalidateRect(itemhWnd, NULL, TRUE);
	ReleaseDC(itemhWnd, itemDC);
	//gColor = color;

	c.rgbtRed = GetRValue(gColor);
	c.rgbtGreen = GetGValue(gColor);
	c.rgbtBlue = GetBValue(gColor);
	b = (gRGBMode == 0) ? 10 : 16;
	if ((excl & 0xFF0) != MASK_RGB)
	{
		itoa(c.rgbtRed, buf, b);
		strupr(buf);
		SetDlgItemText(gHWND, EDIT_RED, buf);
		itoa(c.rgbtGreen, buf, b);
		strupr(buf);
		SetDlgItemText(gHWND, EDIT_GREEN, buf);
		itoa(c.rgbtBlue, buf, b);
		strupr(buf);
		SetDlgItemText(gHWND, EDIT_BLUE, buf);
	}
	if ((excl & 0xFF0) != MASK_PERCENT_RGB)
	{
		gcvt( ((double)c.rgbtRed * 100.0)/255.0 ,4, buf);
		SetDlgItemText(gHWND, EDIT_PERCENT_RED, buf);
		gcvt( ((double)c.rgbtGreen * 100.0)/255.0 , 4, buf);
		SetDlgItemText(gHWND, EDIT_PERCENT_GREEN, buf);
		gcvt( ((double)c.rgbtBlue * 100.0)/255.0 , 4, buf);
		SetDlgItemText(gHWND, EDIT_PERCENT_BLUE, buf);
	}

	if (gHSVorHSLMode == 0)
		rgb_to_hsv(gColor, &h, &s, &v);
	else
		ColorRGBToHLS(gColor, &h, &v, &s);
	b = (gHueMode == 0) ? 10 : 16;
	if ((excl & 0xFF0) != MASK_HUE)
	{
		itoa(h, buf, b);
		strupr(buf);
		SetDlgItemText(gHWND, EDIT_HUE, buf);
		itoa(s, buf, b);
		strupr(buf);
		SetDlgItemText(gHWND, EDIT_SATURATION, buf);
		itoa(v, buf, b);
		strupr(buf);
		SetDlgItemText(gHWND, EDIT_VALUE, buf);
	}
	if ((excl & 0xFF0) != MASK_PERCENT_HUE)
	{
		if (gHuePercent == 0)
			gcvt( ((double)h*360.0)/240.0 , 4, buf);
		else
			gcvt( ((double)h*100.0)/240.0 , 4, buf);
		SetDlgItemText(gHWND, EDIT_PERCENT_HUE, buf);
		gcvt( ((double)s*100.0)/240.0 , 4, buf);
		SetDlgItemText(gHWND, EDIT_PERCENT_SATURATION, buf);
		gcvt( ((double)v*100.0)/240.0 , 4, buf);
		SetDlgItemText(gHWND, EDIT_PERCENT_VALUE, buf);
	}

	if (gCTable != NULL && excl != EDIT_INDEX)
	{
		b = GetNearestPaletteIndex(gCTable, gColor);
		itoa(b, buf, 10);
		SetDlgItemText(gHWND, EDIT_INDEX, buf);
	}

	gUpdating = false;
}

bool CColor::setColorTableIndex(DWORD i)
{
	PALETTEENTRY p;
	if (gUpdating || !gCTable) return false;
	GetPaletteEntries(gCTable, i, 1, &p);
	return setColorRGB(RGB(p.peRed,p.peGreen,p.peBlue), EDIT_INDEX);
}

COLORREF CColor::hsv_to_rgb(WORD h, WORD s, WORD v)
{
	DOUBLE H, S, V, R, G, B;
	DOUBLE p1, p2, p3;
	DOUBLE f;
	DWORD i;

	//if (h < 0) h = 0;
	//if (s < 0) s = 0;
	//if (v < 0) v = 0;
	if (h > 240) h = 240;
	if (s > 240) s = 240;
	if (v > 240) v = 240;

	S = s / 240.0; V = v / 240.0;
	//H = (h % 360) / 60.0;
	H = h / 40.0;
	i = static_cast<DWORD>(H);
	f = H - i;
	p1 = V * (1 - S);
	p2 = V * (1 - (S * f));
	p3 = V * (1 - (S * (1 - f)));
	if      (i == 0) { R = V;  G = p3; B = p1; }
	else if (i == 1) { R = p2; G = V;  B = p1; }
	else if (i == 2) { R = p1; G = V;  B = p3; }
	else if (i == 3) { R = p1; G = p2; B = V;  }
	else if (i == 4) { R = p3; G = p1; B = V;  }
	else             { R = V;  G = p1; B = p2; }
	return RGB(R*255,G*255,B*255);
}

void CColor::rgb_to_hsv(COLORREF rgb, WORD *h, WORD *s, WORD *v)
{
	DOUBLE R, G, B, H, S, V;
	DOUBLE cmax, cmin;
	DOUBLE cmm;
	DWORD imax;
	R = ((DOUBLE) GetRValue(rgb)) / 255.0;
	G = ((DOUBLE) GetGValue(rgb)) / 255.0;
	B = ((DOUBLE) GetBValue(rgb)) / 255.0;
	cmax = R; cmin = G; imax = 1;
	if  ( cmax < G ) { cmax = G; cmin = R; imax = 2; }
	if  ( cmax < B ) { cmax = B; imax = 3; }
	if  ( cmin > B ) { cmin = B; }
	cmm = cmax - cmin;
	V = cmax;
	if (cmm == 0)
	S = H = 0;
	else
	{
		S = cmm / cmax;
		if       (imax == 1)    H =       (G - B) / cmm;
		else  if (imax == 2)    H = 2.0 + (B - R) / cmm;
		else /*if (imax == 3)*/ H = 4.0 + (R - G) / cmm;
		if (H < 0) H += 6.0;
	}
	*h = static_cast<WORD>(H * 40);
	*s = static_cast<WORD>(S * 240);
	*v = static_cast<WORD>(V * 240);
}

